package com.jz.agent.client.service;


import com.google.gson.Gson;
import com.jz.agent.client.DmpDsAgentClient;
import com.jz.agent.config.DmpDsAgentConstants;
import com.jz.dmp.agent.DmpAgentResult;
import com.jz.dmp.agent.ResultCode;
import com.jz.dmp.modules.model.DmpAgentDatasourceInfo;
import org.apache.commons.lang3.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.*;
import org.apache.hadoop.io.IOUtils;
import org.apache.hadoop.security.UserGroupInformation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;

// HDFS没有提供界面操作接口
public class DmpHdfsAgentClientService implements DmpDsAgentClient {
	
	private Gson gson = new Gson();
	
	private static Logger logger = LoggerFactory.getLogger(DmpHdfsAgentClientService.class);

	@SuppressWarnings("unused")
	public DmpAgentResult testConnect(DmpAgentDatasourceInfo ds) {
		FileSystem fileSystem = null;
		try {
			/*String defaultFs = ds.getDefaultFs();
			if(StringUtils.isBlank(defaultFs) && StringUtils.isNotBlank(ds.getJdbcUrl())) {
				defaultFs = ds.getJdbcUrl();
			}*/
			//Configuration configuration = initConfiguration(this.getHadoopConfFilePath(defaultFs));
			//if(DmpDsAgentConstants.IS_ENABLED_KERBEROS) JDBCUtils.initKerberosENV4HDFS(configuration);
			Configuration configuration = setHdfsConfiguration(ds);
			if(configuration == null) {
				return new DmpAgentResult(ResultCode.PARAMS_ERROR, "HDFS参数有误");
			}
			fileSystem = FileSystem.get(configuration);
			fileSystem.getUsed();
			return new DmpAgentResult(ResultCode.SUCCESS, "true");
		}  catch (IOException e) {
			e.printStackTrace();
			return new DmpAgentResult(ResultCode.EXCEPTION, "测试连通性出错："+e.getMessage());
		} finally {
			if(fileSystem != null)
				try {
					fileSystem.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
	}
	
	private Configuration setHdfsConfiguration(DmpAgentDatasourceInfo ds) {
		logger.info("DmpAgentDatasourceInfo: "+gson.toJson(ds));
		Configuration conf = new Configuration();
		//conf.set("hadoop.security.authentication",ds.getHdfsUserName());
		conf.set("fs.hdfs.impl", "org.apache.hadoop.hdfs.DistributedFileSystem");
		String defaultFs = ds.getDefaultFs();
		if(StringUtils.isBlank(defaultFs) && StringUtils.isNotBlank(ds.getJdbcUrl())) {
			defaultFs = ds.getJdbcUrl();
		}
		conf.set("fs.defaultFS", defaultFs);
		logger.info("fs.defaultFS: "+defaultFs);
		if(ds.getKerberosIsenable() != null && "true".equals(ds.getKerberosIsenable())) {
			conf.set("hadoop.security.authentication","kerberos");
			System.setProperty("java.security.krb5.conf",ds.getKerberosKrb5Conf());
			//System.setProperty("java.security.auth.login.config",ds.getKerberosJaasConf());
			
			logger.info("hadoop.security.authentication: "+"kerberos");
			logger.info("java.security.krb5.conf: "+ds.getKerberosKrb5Conf());
		}
		UserGroupInformation.setConfiguration(conf);
		logger.info("kerberos.is.enabled: "+ds.getKerberosIsenable());
		if(ds.getKerberosIsenable() != null && "true".equals(ds.getKerberosIsenable())) {
			try {
				logger.info("hdfs.user.name: "+ds.getHdfsUserName());
				logger.info("hdfs.auth.path: "+ds.getHdfsAuthPath());
				UserGroupInformation.loginUserFromKeytab(ds.getHdfsUserName(), ds.getHdfsAuthPath());
			} catch (IOException e) {
				e.printStackTrace();
				return null;
			}
		}
		return conf;
	}

	public DmpAgentResult getTableNameList(DmpAgentDatasourceInfo ds) {
		List<String> list = new ArrayList<String>();
		FileSystem fileSystem = null;
		try {
			String defaultFs = ds.getDefaultFs();
			if(StringUtils.isBlank(defaultFs) && StringUtils.isNotBlank(ds.getJdbcUrl())) {
				defaultFs = ds.getJdbcUrl();
			}
			//Configuration configuration = initConfiguration(this.getHadoopConfFilePath(defaultFs));
			//if(DmpDsAgentConstants.IS_ENABLED_KERBEROS) JDBCUtils.initKerberosENV4HDFS(configuration);
			Configuration configuration = setHdfsConfiguration(ds);
			if(configuration == null) {
				return new DmpAgentResult(ResultCode.PARAMS_ERROR, "HDFS参数有误");
			}
			fileSystem = FileSystem.get(configuration);
			String filePath = ds.getTargetFileName();
			Path srcPath = new Path(filePath);
			if(fileSystem.exists(srcPath) && fileSystem.isDirectory(srcPath)) {
				RemoteIterator<LocatedFileStatus> ri = fileSystem.listFiles(srcPath, true);
				while(ri.hasNext()) {
					LocatedFileStatus lfs = ri.next();
					System.out.println(lfs.getPath().toString());
					list.add(lfs.getPath().toString());
				}
			}
		}  catch (IOException e) {
			e.printStackTrace();
			return new DmpAgentResult(ResultCode.EXCEPTION, "HDFS获取列表出错："+e.getMessage());
		} finally {
			if(fileSystem != null) {
				try {
					fileSystem.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		return new DmpAgentResult(ResultCode.SUCCESS, gson.toJson(list));
	}

	@SuppressWarnings("unchecked")
	public DmpAgentResult getTableColumnList(DmpAgentDatasourceInfo ds, String targetName) {
		Object o = this.getColumnAndResult(ds, targetName);
		//List<Map<String, Object>> retMap = (List<Map<String, Object>>) map.get("column");
		//return new DmpAgentResult(ResultCode.SUCCESS, gson.toJson(retMap));
		if(o instanceof DmpAgentResult) {
			return new DmpAgentResult(ResultCode.EXCEPTION, gson.toJson(o));
		} else {
			Map<String, Object> map = (Map<String, Object>) o;
			List<Map<String, Object>> retMap = (List<Map<String, Object>>) map.get("column");
			return new DmpAgentResult(ResultCode.SUCCESS, gson.toJson(retMap));
		}
	}

	public DmpAgentResult previewData(DmpAgentDatasourceInfo ds, String targetName) {
		//Map<String, Object> m = this.getColumnAndResult(ds, targetName);
		//m.remove("column");
		//return new DmpAgentResult(ResultCode.SUCCESS, gson.toJson(m));
		Object o = this.getColumnAndResult(ds, targetName);
		if(o instanceof DmpAgentResult) {
			return new DmpAgentResult(ResultCode.EXCEPTION, gson.toJson(o));
		} else {
			Map<String, Object> map = (Map<String, Object>) o;
			map.remove("column");
			return new DmpAgentResult(ResultCode.SUCCESS, gson.toJson(map));
		}
	}
	
	@SuppressWarnings("unchecked")
	private Map<String, Object> getColumnAndResult(DmpAgentDatasourceInfo ds, String filePath) {
		Map<String, Object> map = new HashMap<String, Object>();
		InputStream is = null;
		FileSystem fileSystem = null;
		Path srcPath = new Path(filePath);
		String splitSymbol = ds.getDelimiter();
		String fileType = ds.getFileType();
		int columnNum = 0;
		try {
			String defaultFs = ds.getDefaultFs();
			if(StringUtils.isBlank(defaultFs) && StringUtils.isNotBlank(ds.getJdbcUrl())) {
				defaultFs = ds.getJdbcUrl();
			}
			//Configuration configuration = initConfiguration(this.getHadoopConfFilePath(defaultFs));
			//if(DmpDsAgentConstants.IS_ENABLED_KERBEROS) JDBCUtils.initKerberosENV4HDFS(configuration);
			Configuration configuration = setHdfsConfiguration(ds);
			if(configuration == null) {
				return map;
			}
			fileSystem = FileSystem.get(configuration);
			is = fileSystem.open(srcPath);
			//IOUtils.copyBytes(is, System.out, 4096, false);
			StringBuilder sb = new StringBuilder();
			int rowCnt = 0;
			int rowNum = 6;
			byte[] bs = new byte[1024*50];
            int len = -1;
            List<String> header = new ArrayList<String>();
            List<String> type = new ArrayList<String>();
            while((len = is.read(bs)) != -1){
                sb.append(new String(bs,0,len));
                String s = sb.toString();
                List<Object[]> list = new ArrayList<Object[]>();
                boolean flag = true;
                String enterLine = "\n";
                if(s.indexOf("\r\n") > -1) {
                	enterLine = "\r\n";
                } else if(s.indexOf("\r") > -1) {
                	enterLine = "\r";
                } else if(s.indexOf("\n") > -1) {
                	enterLine = "\n";
                } else {
                	flag = false;
                }
                while(rowCnt < rowNum && flag) {
	                int i1 = s.indexOf(enterLine);
	                String tmp = s.substring(0, i1);
	                if("json".equalsIgnoreCase(fileType)) {
	                	Map<String, Object> m = gson.fromJson(tmp, Map.class);
	                	if(rowCnt == 0) {
	                		for(Map.Entry<String, Object> me : m.entrySet()) {
	                			header.add(me.getKey());
	                			type.add(this.getObjectType(me.getValue()));
	                			columnNum++;
	                		}
	                	}
	                	Object[] row = new Object[columnNum];
	                	int i = 0;
	                	for(String rn : header) {
	                		row[i] = m.get(rn);
	                		i++;
	                	}
	                	list.add(row);
	                } else {
	                	if(StringUtils.isBlank(splitSymbol)) {
		                	if(tmp.indexOf(",") > -1) {
			                	splitSymbol = ",";
			                } else if(tmp.indexOf("\t") > -1) {
			                	splitSymbol = "\t";
			                } else if(tmp.indexOf(" ") > -1) {
			                	splitSymbol = " ";
			                }
		                }
		                String[] tArr = tmp.split(splitSymbol);
		                list.add(tArr);
		                columnNum = tArr.length;
	                }
	                s = s.substring(i1+2, s.length());
	                rowCnt++;
                }
                if(rowCnt == rowNum) {
                	map.put("result", list);
                	break;
                } else {
                	rowCnt = 0;
					rowNum = 6;
                }
            }
            List<Map<String, Object>> l2 = new ArrayList<Map<String, Object>>();
            List<String> l3 = new ArrayList<String>();
            if("json".equalsIgnoreCase(fileType)) {
            	for(int i = 0; i < header.size(); i++) {
            		Map<String, Object> m = new HashMap<String, Object>();
            		m.put("key", i);
        			m.put("name", header.get(i));
        			m.put("type", type.get(i));
        			l2.add(m);
        			l3.add(header.get(i).toUpperCase());
            	}
            } else {
            	if("1".equals(ds.getIsHaveHeader())){
                	List<String[]> l = (List<String[]>) map.get("result");
                	String[] arr = l.get(0);
                	for(int i = 0; i < columnNum; i++) {
            			Map<String, Object> m = new HashMap<String, Object>();
            			m.put("key", i);
            			m.put("name", arr[i]);
            			m.put("type", "STRING");
            			l2.add(m);
            			l3.add(arr[i].toUpperCase());
            		}
                	map.put("result", l.subList(1, l.size()));
                } else {
                	for(int i = 0; i < columnNum; i++) {
            			Map<String, Object> m = new HashMap<String, Object>();
            			m.put("key", i);
            			m.put("name", "第"+i+"列");
            			m.put("type", "STRING");
            			l2.add(m);
            			l3.add("第"+i+"列");
            		}
                }
            }
            map.put("header", l3);
            map.put("column", l2);
		}  catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			if(fileSystem != null) {
				try {
					fileSystem.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		return map;
	}
	
	private String getObjectType(Object obj) {
		String t = null;
		if(obj instanceof Long) {
			t = "LONG";
		} else if(obj instanceof Double) {
			t = "DOUBLE";
		} else if(obj instanceof Boolean) {
			t = "BOOLEAN";
		} else if(obj instanceof Date) {
			t = "DATE";
		} else {
			t = "STRING";
		}
		return t;
	}
	
	public String getHadoopConfFilePath(String nameNodeAddr) {
		String defaultFs = nameNodeAddr;
		defaultFs = defaultFs.replace("hdfs://", "");
		int idx = defaultFs.indexOf(":");
		String hadoopConfIp = defaultFs.substring(0, idx);
		return DmpDsAgentConstants.HADOOP_CONF_FILE_PATH + hadoopConfIp + "/hadoop-conf";
	}

	/**
	 * 将该任务对应的表数据转换为xml文件 并上传到hdfs指定路径（/jz/etlprocesses/guide_xml/目标库名/）
	 * Xml名称：目标库名_目标表名（等同于xml文件中name标签的值）
	 */
	public DmpAgentResult submitSycingConfig(DmpAgentDatasourceInfo ds, String xml) {
		FileSystem fileSystem = null;
		try {
			//Configuration configuration = initConfiguration(DmpDsAgentConstants.HADOOP_DEFAULT_CONF_FILE_PATH);
			//if(DmpDsAgentConstants.IS_ENABLED_KERBEROS) JDBCUtils.initKerberosENV4HDFS(configuration);
			Configuration configuration = setHdfsConfiguration(ds);
			if(configuration == null) {
				return new DmpAgentResult(ResultCode.PARAMS_ERROR, "HDFS参数有误");
			}
			fileSystem = FileSystem.get(configuration);
			/*
			mkdir(fileSystem, DmpDsAgentConstants.SUBMIT_SYNCING_CONFIG_HDFS_PATH 
					+ ds.getTargetDbName() + "/");
			createFile(fileSystem, DmpDsAgentConstants.SUBMIT_SYNCING_CONFIG_HDFS_PATH 
					+ ds.getTargetDbName() + "/" + ds.getTargetFileName(), xml);
			*/
			String hasSeparate = ds.getHdfsSyncingPath().substring(ds.getHdfsSyncingPath().length() - 1);
			String pathSup = hasSeparate.equals("/") ? "" : "/";
			String mkdirPath = ds.getHdfsSyncingPath() + pathSup + ds.getProjectId() + "/" + ds.getTargetDbName() + "/";
			logger.info("mkdir: " + mkdirPath);
			mkdir(fileSystem, mkdirPath);
			logger.info("createFile : " + mkdirPath + ds.getTargetFileName());
			logger.info("file content : " + xml);
			createFile(fileSystem, mkdirPath + ds.getTargetFileName(), xml);
			return new DmpAgentResult(ResultCode.SUCCESS, "true");
		} catch (IOException e) {
			e.printStackTrace();
			return new DmpAgentResult(ResultCode.EXCEPTION, "上传出错："+e.getMessage());
		} finally {
			if(fileSystem != null)
				try {
					fileSystem.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
	}

	public static Configuration initConfiguration(String confPath) {
		Configuration configuration = new Configuration();
		System.out.println(confPath + File.separator + "core-site.xml");
		configuration.addResource(new Path(confPath + File.separator + "core-site.xml"));
		configuration.addResource(new Path(confPath + File.separator + "hdfs-site.xml"));
		configuration.set("fs.hdfs.impl", "org.apache.hadoop.hdfs.DistributedFileSystem");
		System.setProperty("HADOOP_USER_NAME", DmpDsAgentConstants.AGENT_HADOOP_USER_NAME);
		return configuration;
	}

	public static void mkdir(FileSystem fs, String path) {
		try {
			Path srcPath = new Path(path);
			if (fs.exists(srcPath)) {
				System.out.println("目录已存在");
				return;
			}
			boolean isok = fs.mkdirs(srcPath);
			if (isok) {
				System.out.println("create dir ok!");
			} else {
				System.out.println("create dir failure");
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static void uploadFile(FileSystem fs, String src, String dst) {
		try {
			Path srcPath = new Path(src);
			Path dstPath = new Path(dst);
			fs.copyFromLocalFile(false, srcPath, dstPath);
			FileStatus[] fileStatus = fs.listStatus(dstPath);
			for (FileStatus file : fileStatus) {
				System.out.println(file.getPath());
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static void createFile(FileSystem fs, String dst, String contents) {
		try {
			Path path = new Path(dst);
			FSDataOutputStream fsDataOutputStream = fs.create(path);
			fsDataOutputStream.write(contents.getBytes());
			fsDataOutputStream.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public static void readFile(FileSystem fs, String filePath) {
        try {
            Path srcPath = new Path(filePath);
            InputStream in = null;
            in = fs.open(srcPath);
            IOUtils.copyBytes(in, System.out, 4096, false); //复制到标准输出流

        } catch (IOException e) {
            e.printStackTrace();
        }
    }

}
